/* jabberd - a dispatcher program for jabber 2.x
   Copyright (C) 2007 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301 USA. */

#include "jabberd.h"

struct prog
{
  struct prog *next;
  pid_t pid;
  char *cmd;
  char *cfg;
  time_t timestamp;
  size_t count;
  int disabled;
};


static struct prog *proghead, *progtail;
static size_t prognum;

void
register_prog (char *cmd, char *cfg)
{
  size_t cmdlen = strlen (cmd) + 1;
  size_t cfglen = cfg ? (strlen (cfg) + 1) : 0;
  struct prog *pp = emalloc (sizeof (*pp) + cmdlen + cfglen);
  memset(pp, 0, sizeof(*pp));
  pp->pid = 0;
  pp->cmd = (char *) (pp + 1);
  strcpy (pp->cmd, cmd);
  if (cfglen)
    {
      pp->cfg = (char *) (pp + 1) + cmdlen + 1;
      strcpy (pp->cfg, cfg);
    }
  else
    pp->cfg = NULL;
  pp->next = NULL;
  if (progtail)
    progtail->next = pp;
  else
    proghead = pp;
  progtail = pp;
}

static struct prog *
find_prog (pid_t pid)
{
  struct prog *prog;

  for (prog = proghead; prog; prog = prog->next)
    if (prog->pid == pid)
      break;
  return prog;
}

size_t
progman_running_count ()
{
  size_t size = 0;
  struct prog *prog;
    
  for (prog = proghead; prog; prog = prog->next)
    if (prog->pid > 0)
      size++;
  return size;
}

static void
prog_start (struct prog *prog)
{
  int i;
  pid_t pid;
  char *argv[5];
  int argc = 0;
  time_t now;
  
  argv[argc++] = prog->cmd;
  if (prog->cfg)
    {
      argv[argc++] = "-c";
      argv[argc++] = prog->cfg;
    }
  if (debug_level > 2)
    argv[argc++] = "-D";
  argv[argc] = NULL;

  time (&now);

  if (prog->timestamp + TESTTIME > now)
    prog->count++;
  else
    {
      prog->count = 0;
      prog->timestamp = now;
    }

  if (prog->count > MAXSPAWN)
    {
      int old_alarm;
      
      logmsg(LOG_ERR, "%s is respawning too fast, disabled for %d minutes",
	     prog->cmd, SLEEPTIME / 60);
      prog->timestamp = now;
      prog->disabled = 1;
      
      old_alarm = alarm (0);
      if (old_alarm > SLEEPTIME || old_alarm <= 0)
	old_alarm = SLEEPTIME;
      alarm (old_alarm);
      return;
    }
    
  logmsg (LOG_DEBUG, "starting %s", prog->cmd);
  
  switch (pid = fork ())
    {
      /* The child branch.  */
    case 0:
       /* Close unneded descripitors */
       for (i = getmaxfd (); i > 2; i--)
	 close (i);

       signal_setup (SIG_DFL);

       execvp(argv[0], argv);
       openlog (syslog_tag, LOG_PID, log_facility);
       logmsg (LOG_CRIT, "cannot start `%s': %s", prog->cmd,
	       strerror (errno));
       exit (1);

    case -1:
      logmsg (LOG_CRIT, "cannot runt `%s': fork failed: %s",
	      prog->cmd, strerror (errno));
      break;

    default:
      prog->pid = pid;
    }
}

void
progman_start ()
{
  struct prog *prog;

  for (prog = proghead; prog; prog = prog->next)
    prog_start (prog);
}

void
progman_wake_disabled ()
{
  struct prog *prog;

  for (prog = proghead; prog; prog = prog->next)
    if (prog->disabled)
      {
	prog->disabled = 0;
	prog->count = 0;
	prog->timestamp = 0;
	prog_start (prog);
      }
}

static void
prog_stop_recursive (struct prog *prog, int sig)
{
  if (!prog)
    return;
  prog_stop_recursive (prog->next, sig);
  if (prog->pid > 0)
    {
      logmsg (LOG_DEBUG, "Stopping %s",
	      prog->cmd, (unsigned long) prog->pid);
      kill (prog->pid, sig);
    }
}

void
progman_stop ()
{
  int i;
  
  prog_stop_recursive (proghead, SIGTERM);
  for (i = 0; i < 3; i++)
    {
      progman_cleanup (1);
      if (progman_running_count () == 0)
	return;
      sleep (1);
    }
  prog_stop_recursive (proghead, SIGKILL);
}

static void
print_status (struct prog *prog, int status, int expect_term)
{
  if (WIFEXITED (status))
    {
      if (WEXITSTATUS (status) == 0)
	logmsg (LOG_DEBUG,
		"%s (%lu) exited successfully",
		prog->cmd, (unsigned long) prog->pid);
      else
	logmsg (LOG_ERR,
		"%s (%lu) failed with status %d",
		prog->cmd, (unsigned long) prog->pid,
		WEXITSTATUS (status));
    }
  else if (WIFSIGNALED (status))
    {
      int prio;

      if (expect_term && WTERMSIG (status) == SIGTERM)
	prio = LOG_DEBUG;
      else
	prio = LOG_ERR;

      logmsg(prio,
	     "%s (%lu) terminated on signal %d",
	     prog->cmd, (unsigned long) prog->pid,
	     WTERMSIG (status));
    }
  else if (WIFSTOPPED (status))
    logmsg(LOG_ERR,
	   "%s (%lu) stopped on signal %d",
	   prog->cmd, (unsigned long) prog->pid,
	   WSTOPSIG (status));
#ifdef WCOREDUMP
  else if (WCOREDUMP (status))
    logmsg (LOG_ERR,
	    "%s (%lu) dumped core",
	    prog->cmd, (unsigned long) prog->pid);
#endif
  else
    logmsg(LOG_ERR,
	   "%s (%lu) terminated with unrecognized status",
	   prog->cmd, (unsigned long) prog->pid);
}

void
progman_cleanup (int expect_term)
{
  pid_t pid;
  int status;
  while ((pid = waitpid (-1, &status, WNOHANG)) > 0)
    {
      struct prog *prog = find_prog (pid);
      if (!prog)
	logmsg (LOG_NOTICE, "unknown child %lu finished",
		(unsigned long) pid);
      print_status (prog, status, expect_term);
      prog->pid = 0;
      if (!expect_term)
	prog_start (prog);
    }
}
